#include <stdio.h>
#include <stdlib.h>

struct tbl {
  float val;
  unsigned long idx;
};

enum errortype { none=0, order=1, range=2, match=3 };

void __cdecl sort4(struct tbl *ptr);
void __cdecl sort8(struct tbl *ptr);
void __cdecl sort12(struct tbl *ptr);

char *failure[] = {"none", "order", "range", "match"};

struct tbl datatbl[1000], savetbl[1000];

struct tbl data2tbl[] = {
    {-6.0, 1},
    {0.0, 2},
    {-1.0, 3},
    {3.0, 4}
  };

struct tbl data3tbl[] = {
    {6.0, 1},
    {3.0, 2},
    {4.0, 3},
    {4.0, 4},
    {9.0, 5},
    {1.0, 6},
    {5.0, 7},
    {2.0, 8}
};
struct tbl data4tbl[] = {
    {-6.0, 1},
    {3.0, 2},
    {-1.0, 3},
    {9.0, 4},
    {4.0, 5},
    {1.0, 6},
    {-11.0, 7},
    {0.0, 8}
  };


int main (int argv, char **argc)
{
  int k, j, errcnt;
  float roff;
  enum errortype error;

  errcnt = 0;

  // Test the 4 entry sort function

  for (j = 0; j < 1000; j++) {
    roff = rand();
    for (k = 0; k < 4; k++) {
      datatbl[k].val = (rand() - roff) * 0.0001;
      datatbl[k].idx = k;
      savetbl[k] = datatbl[k];
    }
#ifdef DEBUG
    printf("Sort: {%-8.3g, %-8.3g, %-8.3g, %-8.3g)\n",
	   datatbl[0].val, datatbl[1].val, datatbl[2].val, datatbl[3].val);
#endif

    sort4(datatbl);

    error = none;
    for (k = 0; k < 3; k++) {
      if (datatbl[k].val < datatbl[k+1].val)
	error = order;
      else if (datatbl[k].idx > 3 || datatbl[k].idx < 0)
	error = range;
      else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
	error = match;
    }
    if (datatbl[k].idx > 3 || datatbl[k].idx < 0)
      error = range;
    else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
      error = match;

    if (error != none) {
      errcnt++;
      printf("\nOriginal datatbl[%u]:\n", j);
      for (k = 0; k < 4; k++)
	printf("v=%-6.2f, idx=%lu\n", savetbl[k].val, savetbl[k].idx);
      printf("Sorted (err=%s):\n", failure[error]);
      for (k = 0; k < 4; k++)
	printf("v=%-6.2f, idx=%lu\n", datatbl[k].val, datatbl[k].idx);
    }
  }

  for (k = 0; k < 4; k++) {
    data2tbl[k].idx = k;
    savetbl[k] = datatbl[k] = data2tbl[k];

  }
  
  sort4(datatbl);

  error = none;
  for (k = 0; k < 3; k++) {
    if (datatbl[k].val < datatbl[k+1].val)
      error = order;
    else if (datatbl[k].idx > 3 || datatbl[k].idx < 0)
      error = range;
    else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
      error = match;
  }
  if (datatbl[k].idx > 3 || datatbl[k].idx < 0)
    error = range;
  else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
    error = match;

  if (error != none) {
    errcnt++;
    printf("\nOriginal data2tbl:\n");
    for (k = 0; k < 4; k++)
      printf("v=%-6.2f, idx=%lu\n", savetbl[k].val, savetbl[k].idx);
    printf("Sorted (err=%s):\n", failure[error]);
    for (k = 0; k < 4; k++)
      printf("v=%-6.2f, idx=%lu\n", datatbl[k].val, datatbl[k].idx);
  }

  //===================================================================

  // Test the 8 entry sort function

  for (j = 0; j < 1000; j++) {
    roff = rand();
    for (k = 0; k < 8; k++) {
      datatbl[k].val = (rand() - roff) * 0.0001;
      datatbl[k].idx = k;
      savetbl[k] = datatbl[k];
    }
#ifdef DEBUG
    printf("Sort: {%-8.3g, %-8.3g, %-8.3g, %-8.3g, %-8.3g, %-8.3g, %-8.3g, %-8.3g)\n",
	   datatbl[0].val, datatbl[1].val, datatbl[2].val, datatbl[3].val
	   datatbl[4].val, datatbl[5].val, datatbl[6].val, datatbl[7].val);
#endif

    sort8(datatbl);

    error = none;
    for (k = 0; k < 7; k++) {
      if (datatbl[k].val < datatbl[k+1].val)
	error = order;
      else if (datatbl[k].idx > 7 || datatbl[k].idx < 0)
	error = range;
      else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
	error = match;
    }
    if (datatbl[k].idx > 7 || datatbl[k].idx < 0)
      error = range;
    else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
      error = match;

    if (error != none) {
      errcnt++;
      printf("\nOriginal datatbl[%u]:\n", j);
      for (k = 0; k < 8; k++)
	printf("v=%-6.2f, idx=%lu\n", savetbl[k].val, savetbl[k].idx);
      printf("Sorted (err=%s):\n", failure[error]);
      for (k = 0; k < 8; k++)
	printf("v=%-6.2f, idx=%lu\n", datatbl[k].val, datatbl[k].idx);
    }
  }

  for (k = 0; k < 8; k++) {
    data3tbl[k].idx = k;
    savetbl[k] = datatbl[k] = data3tbl[k];
  }

  sort8(datatbl);

  error = none;
  for (k = 0; k < 7; k++) {
    if (datatbl[k].val < datatbl[k+1].val)
      error = order;
    else if (datatbl[k].idx > 7 || datatbl[k].idx < 0)
      error = range;
    else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
      error = match;
  }
  if (datatbl[k].idx > 7 || datatbl[k].idx < 0)
    error = range;
  else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
    error = match;

  if (error != none) {
    errcnt++;
    printf("\nOriginal data3tbl:\n");
    for (k = 0; k < 8; k++)
      printf("v=%-6.2f, idx=%lu\n", savetbl[k].val, savetbl[k].idx);
    printf("Sorted (err=%s):\n", failure[error]);
    for (k = 0; k < 8; k++)
      printf("v=%-6.2f, idx=%lu\n", datatbl[k].val, datatbl[k].idx);
  }

  for (k = 0; k < 8; k++) {
    data4tbl[k].idx = k;
    savetbl[k] = datatbl[k] = data4tbl[k];
  }

  sort8(datatbl);

  error = none;
  for (k = 0; k < 7; k++) {
    if (datatbl[k].val < datatbl[k+1].val)
      error = order;
    else if (datatbl[k].idx > 7 || datatbl[k].idx < 0)
      error = range;
    else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
      error = match;
  }
  if (datatbl[k].idx > 7 || datatbl[k].idx < 0)
    error = range;
  else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
    error = match;

  if (error != none) {
    errcnt++;
    printf("\nOriginal data4tbl:\n");
    for (k = 0; k < 8; k++)
      printf("v=%-6.2f, idx=%lu\n", savetbl[k].val, savetbl[k].idx);
    printf("Sorted (err=%s):\n", failure[error]);
    for (k = 0; k < 8; k++)
      printf("v=%-6.2f, idx=%lu\n", datatbl[k].val, datatbl[k].idx);
  }

  //===================================================================

  // Test the 12 entry sort function


  for (j = 0; j < 1000; j++) {
    roff = rand();
    for (k = 0; k < 12; k++) {
      datatbl[k].val = (rand() - roff) * 0.0001;
      datatbl[k].idx = k;
      savetbl[k] = datatbl[k];
    }
#ifdef DEBUG
    printf("Sort: {%-8.3g, %-8.3g, %-8.3g, %-8.3g, %-8.3g, %-8.3g,\n       "
	   "%-8.3g, %-8.3g,  %-8.3g, %-8.3g, %-8.3g, %-8.3g\n",
	   datatbl[0].val, datatbl[1].val, datatbl[2].val, datatbl[3].val,
	   datatbl[4].val, datatbl[5].val, datatbl[6].val, datatbl[7].val,
	   datatbl[8].val, datatbl[9].val, datatbl[10].val, datatbl[11].val);
#endif

    sort12(datatbl);

    error = none;
    for (k = 0; k < 11; k++) {
      if (datatbl[k].val < datatbl[k+1].val)
	error = order;
      else if (datatbl[k].idx > 11 || datatbl[k].idx < 0)
	error = range;
      else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
	error = match;
    }
    if (datatbl[k].idx > 11 || datatbl[k].idx < 0)
      error = range;
    else if (savetbl[datatbl[k].idx].val != datatbl[k].val)
      error = match;

    if (error != none) {
      errcnt++;
      printf("\nOriginal datatbl[%u]:\n", j);
      for (k = 0; k < 12; k++)
	printf("v=%-6.2f, idx=%lu\n", savetbl[k].val, savetbl[k].idx);
      printf("Sorted (err=%s):\n", failure[error]);
      for (k = 0; k < 12; k++)
	printf("v=%-6.2f, idx=%lu\n", datatbl[k].val, datatbl[k].idx);
    }
  }

  printf("\nTotal sorts in error: %u\n", errcnt);

  return 0;
}
